/***********************************************************************************************************************
 *
 * blueMarine Metadata - open source media workflow
 * Copyright (C) 2007-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~metadata-src
 *
 **********************************************************************************************************************/
package it.tidalwave.metadata.spi;

import org.openide.loaders.DataObject;
import it.tidalwave.metadata.Metadata.StorageType;
import it.tidalwave.metadata.Metadata.StoreOption;
import it.tidalwave.metadata.MetadataItemHolder;
import javax.annotation.Nonnull;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public interface MetadataItemSink<Item>
  {
    /***************************************************************************
     *
     * Returns the {@link Class} representing the metadata item supported by this 
     * sink.
     * 
     * @return  the metadata item class
     * 
     **************************************************************************/
    @Nonnull 
    public Class<Item> getItemClass();
    
    /***************************************************************************
     *
     * Return the type of this sink.
     * 
     * @return  the sink type
     * 
     **************************************************************************/
    @Nonnull 
    public StorageType getType();
    
    /***************************************************************************
     *
     * Return the name of this sink.
     * 
     * @return  the sink name
     * 
     **************************************************************************/
    @Nonnull 
    public String getName();
    
    /***************************************************************************
     *
     * Stores an item for a given {@link DataObject}. 
     * 
     * @param  dataObject  the object to store the item to
     * @param  holder      the holder
     * 
     **************************************************************************/
    public void storeMetadataItem (@Nonnull DataObject dataObject, 
                                   @Nonnull MetadataItemHolder<Item> holder,
                                   @Nonnull StoreOption ... options)
      throws Exception;
  }
